/* Copyright (c) 2010, Nathan Bubna
 * Dual licensed under the MIT and GPL licenses:
 *   http://www.opensource.org/licenses/mit-license.php
 *   http://www.gnu.org/licenses/gpl.html
 */
/*
JSON-S ("JSON Sharing") provides support for retrieving public JSON resources
via either cross origin resource sharing and/or JSONP.  This provides a
consistent interface across browsers and adds many useful features.  This
defaults to CORS on supporting browsers and JSONP on the rest, essentially
assuming that any targetted server supports both (for now).

    // simple usage (defaults to best available transport):
    JSONS("http://public.api.com/json-resource", function(json) {
        console.log('successfully retrieved: ', json);
    });

    // advanced usage:
    JSONS.get({
        url: 'http://api.com/find',     // base url
        data: { q: 'good pizza' },      // append data to url
        jsonp: true,                    // force jsonp usage
        param: 'jsonpCallback',         // change jsonp param from 'callback'
        cache: true,                    // cache results
        
        // register callbacks for all supported events
        before: function(jsonsOpts) {
            console.debug(jsonsOpts, 'before script tag insertion');
        },
        success: function(data, jsonsOpts) {
            console.log(data, 'success!');
        },
        error: function(errorMsg, jsonsOpts) {
            console.error(errorMsg, jsonsOpts);
        },
        after: function(jsonsOpts) {
            console.debug(jsonsOpts, 'after success/error');
        }
    });

You can also set options globally (some are only global!) and register
global event handlers, which are always called before call-specific ones
like those in the previous example.

    // silence logging
    JSONS.log = false;
    // bow to some API's rate limit of 2 calls/second
    JSONS.throttle = 500;
    // adjust jsonp default
    JSONS.options.param = 'jsonpCallback';
    // set a default timeout to abort calls longer than 30s
    JSONS.options.timeout = 30*1000;
    // set headers for a CORS request
    JSONS.options.headers = { Accept: 'application/json' };
    // register global event handlers to display "Loading..." message
    JSONS.on('before', function(){ jQuery.loading(true); });
    JSONS.on('after', function(){ jQuery.loading(false); });

 */
;(function(window, document, undefined, setTimeout, clearTimeout, encodeURIComponent) {

    // main function
    var JSONS = function(){ return JSONS.get.apply(this, arguments); },
        onRSC = 'onreadystatechange',
        head = document.getElementsByTagName("head")[0] || document.documentElement;

    JSONS.version = "1.0";
    // publicize JSONS, but maintain access to any overwritten var
    if (window.JSONS) JSONS.was = window.JSONS;
    window.JSONS = JSONS;

    // options setup function serves as container for default opts
    var opts = JSONS.options = function(optsOrUrl, success) {
        if (typeof optsOrUrl == "string") {
            var o = { url: optsOrUrl };//just url
        } else {
            var o = {};//copy all opts
            for (var opt in optsOrUrl) o[opt] = optsOrUrl[opt];
        }
        o.count = ++JSONS.count;
        for (var opt in opts) if (o[opt] === undefined) o[opt] = opts[opt];
        if (success) o.success = success;
        return o;
    };
    JSONS.count = 0;
    JSONS.nojson = 'JSON did not load';
    JSONS.outoftime = 'Request timed out';
    JSONS.notjsonp = 'Script was not valid JSONP';
    JSONS.log = 'console' in window;
    JSONS.debug = false;
    JSONS.throttle = false;
    JSONS.XHR = 'withCredentials' in new XMLHttpRequest ? XMLHttpRequest :
                'XDomainRequest' in window ? XDomainRequest : undefined;
    opts.jsonp = !JSONS.XHR;
    opts.param = "callback";
    opts.timeout = 60*1000;//1min, for when script.onerror fails (e.g. Opera)
    opts.RSCTimeout = 1000;//give IE 1s to execute callback after script is loaded
    opts.clean = true;
    opts.cache = false;

    JSONS.get = function(optsOrUrl, success) {
        var o = JSONS.options(optsOrUrl, success);
        JSONS.callback(o);
        JSONS.src(o);
        if (o.cache && (o.cached = JSONS.cache(o))) {
            o.callback(o.cached);
        } else {
            JSONS.send(o);
        }
        return o;
    };

    // register handlers for before/success/error/after events
    JSONS.on = function(event, fn) {
        var list = JSONS.on[event];
        if (list) list.push(fn);
        else JSONS.on[event] = [fn];
    };

    JSONS.send = function(o) {
        if (JSONS.throttle) {
            var now = new Date().valueOf(), since = now - (JSONS.last || 0);
            if (since < JSONS.throttle) {
                return setTimeout(function(){ JSONS.send(o); }, JSONS.throttle);
            }
            JSONS.last = now;
        }
        o.jsonp ? JSONS.tag(o) : JSONS.xhr(o);
        if (o.timeout) JSONS.timer(o);
    };

    JSONS.src = function(o) {
        var s = o.url + (o.url.indexOf('?') > 0 ? '&' : '?');
        if (o.data) s = JSONS.addParams(s, o.data);
        if (JSONS.data) s = JSONS.addParams(s, JSONS.data);
        if (o.cache) o.id = s.substring(0, s.length-1);
        if (o.jsonp) s += JSONS.encode(o.param, 'JSONS['+o.count+']');
        return o.src = s;
    };

    JSONS.addParams = function(url, data) {
        for (var key in data) {
            url += JSONS.encode(key, data[key]) + '&';
        }
        return url;
    };

    JSONS.encode = function(key, val) {
        var s = encodeURIComponent(key)+'=';
        if (Object.prototype.toString.call(val) === "[object Array]") {// if array
            var out = '';
            for (var i=0; i<val.length; i++) {
                if (i > 0) out += '&';
                out += s + (encodeURIComponent(val[i])||'');
            }
            return out;
        } else {
            return s + (encodeURIComponent(val)||'');
        }
    };

    JSONS.callback = function(o) {
        o.callback = JSONS[o.count] = function(result) {
            o.called = true;
            o.result = result;
            if (o.cache && !o.cached) JSONS.cache(o);
            event('success', o, [result, o]);
            JSONS.end(o);
        };
    };

    JSONS.xhr = function(o) {
        var x = o.xhr = new JSONS.XHR();
        x.open('get', o.src, true);
        if (o.headers) for (var n in o.headers) x.setRequestHeader(n, o.headers[n]);
        x.onload = function() {
            try {
                o.callback(JSONS.parse(o.xhr.responseText));
            } catch (e) {
                JSONS.end(o, e);
            }
        };
        x.onerror = function(){ JSONS.end(o, JSONS.nojson); };
        event('before', o);
        x.send();
    };

    JSONS.parse = function(res, o) {
        return 'JSON' in window ? JSON.parse(res) : new Function("return "+res)();
    };

    JSONS.tag = function(o) {
        var tag = document.createElement("script");
        tag.src = o.src;
        if (o.type) tag.type = o.type;
        if (o.charset) tag.charset = o.charset;
        tag.async = 'async';
        tag.onload = function(){ if (!o.called) JSONS.end(o, JSONS.notjsonp); };
        tag.onerror = function(){ JSONS.end(o, JSONS.nojson);  };
        tag[onRSC] = function() {
            o.state = o.script.readyState;
            if (o.state === "loaded") {
                o.rsctimer = setTimeout(function() {
                    if (!o.called) JSONS.end(o, JSONS.nojson);
                }, o.RSCTimeout);// wait a sec for callback to run
            } else if (o.state === "complete") {
                clearTimeout(o.rsctimer);
                if (!o.called) JSONS.end(o, JSONS.notjsonp);
            }
        };
        event('before', o);
        setTimeout(function(){ head.insertBefore(tag, head.firstChild); }, 0);
        if (JSONS.log) console.log('JSONP ', o.src, o);
        return o.script = tag;
    };

    JSONS.timer = function(o) {
        o.timer = setTimeout(function(){ JSONS.end(o, JSONS.outoftime); }, o.timeout);
    };

    JSONS.end = function(o, error) {
        if (o.timer) clearTimeout(o.timer);
        event('after', o);
        if (error) event('error', o, [error, o]);
        if (o.clean != false) {
            try {
                JSONS[o.count] = undefined;
                delete JSONS[o.count];
                if (o.script) {
                    o.script.onload = o.script.onerror = o.script[onRSC] = null;
                    head.removeChild(o.script);
                }
                if (o.xhr) o.xhr.onload = o.xhr.onerror = o.xhr[onRSC] = null;
            } catch(e) {}// ignore cleanup errors
        }
    };

    JSONS.cache = function(o) {
        o.cached = JSONS.cache[o.id];
        if (o.result) JSONS.cache[o.id] = o.result;
        return o.cached;
    };

    // execute global and call-specific event handlers
    function event(type, o, args) {
        if (JSONS.debug) console.debug(type, o.src, o, args);
        if (!args) args = [o];
        var handlers = JSONS.on[type], fn;
        if (handlers) {
            for (var i=0; i<handlers.length; i++) {
                fn = handlers[i];
                fn.apply(fn, args);
            }
        }
        fn = o[type];
        if (fn) fn.apply(fn, args);
    }

})(window, document, undefined, setTimeout, clearTimeout, encodeURIComponent);
