/**
 * Copyright (c) 2011, Nathan Bubna
 * Dual licensed under the MIT and GPL licenses:
 *   http://www.opensource.org/licenses/mit-license.php
 *   http://www.gnu.org/licenses/gpl.html
 *
 * Manages auto-syncing of named DOM elements to each other (automatically)
 * and to specified objects (manually specified).
 *
 * @version 0.3
 * @name sync
 * @author Nathan Bubna
 */
;(function($, document, undefined) {
    // use to sync models with DOM for specified name(s)
    var sync = window.sync = function(name, model, fill) {
        if ($.isPlainObject(name)) {
            fill = model; model = name;
            for (name in models) sync(name, model);
        } else if ($.isArray(name)) {
            for (var i=0,m=name.length; i<m; i++) sync(name[i], model);
        } else {
            if (sync.debug) console.log('sync', name, model);
            (models[name] || (models[name] = [])).push(model);// add model to name's list
            if (!fill && !(name in model)) {
                $('[name="'+name+'"]:first').trigger('sync');// catch up to DOM
            }
        }
        if (fill) sync.fill(model);
    };
    // use specified object/model to fill DOM and other models
    sync.fill = function(source, prefix, root) {// prefix/root are for recursion only
        if (!root && sync.debug) console.log('sync.fill', source);
        for (var name in source) {
            var value = source[name],
                name = (prefix||'')+name;
            if (sync.match(name, sync.write)) sync.set(name, value, root || source);
            if ($.isPlainObject(value)) {
                sync.fill(value, name+'.', root || source);
            } else if ($.isArray(value) && value.length) {
                sync.fill(value, name+':', root || source);
            } else {
                sync.set(name, value, root || source);
            }
        }
    };
    // set value for all DOM and models with specified name
    sync.set = function(name, value, source) {
        if (sync.debug) console.log('sync.set', name, value, source);
        var node = $(source && source.nodeName ? source : document);
        sync.dom(name, value, source);
        sync.model(name, value, sync.match(name, sync.models) || [], source);
        node.trigger('synced.'+name, [name, value, source]);
    };

    sync.read = function(node) {
        if (node.is('input,select,textarea')) {
            //TODO: for groups, this will fail.  all nodes of the same name must be read
            //      and their values returned concatenated
            if (node.is(':checkbox,:radio') && !node.is(':checked')) return '';
            var val = node.val();
            return val.join ? val.join(',') : val;
        }
        if (node.is('[value]')) return node.attr('value');
        return sync.toText(node);
    };
    sync.toText = function(node) {
        var txt = '';
        if (node.is(':visible')) {
            if (node.is('input,textarea')) {
                txt = node.val();
            } else if (node.is('select')) {
                txt = node.find(':selected').text();
            } else if (node.is(':has(select)')) {
                node.contents().each(function(i, n){
                    txt += sync.toText($(n));
                    if (txt.charAt(txt.length-1) != ' ') txt += ' ';
                });
            } else {
                txt = node.text();
            }
        }
        return $.trim(txt);
    };
    sync.write = function(node, value) {
        if (node.is('input,select,textarea')) {
            if (node.is('[multiple],:checkbox,:radio')) value = value.split(',');
            node.val(value);
        } else if (node.is('[value]')) {
            node.attr('value', value);
        } else {
            node.html(value);
        }
    };
    sync.dom = function(name, value, source) {
        if (value === undefined || value === null) value = '';
        var nodes = $('[name="'+name+'"]'),
            write = sync.match(name, sync.write);
        if (write) return write(nodes, value, source);
        nodes.each(function(i, node) {
            if (node != source) sync.write($(node), value);
        });
    };
    var models = sync.models = {};// map of full names to lists of models
    sync.model = function(name, value, models, source) {
        var keys = name.split('.'), model, key, index, array, parts;
        for (var j=0,n=models.length; j<n; j++) {
            model = models[j];
            if (model != source) {
                for (var i=0,m=keys.length; i<m; i++) {
                    parts = keys[i].split(':');
                    key = parts[0];
                    index = parts[1];
                    if (index == 'i') index = sync.index($(source));
                    array = index != undefined;
                    if (i == m - 1) {
                        if (array) {
                            if (!model[key]) model[key] = [];
                            model[key][index] = value;
                        } else {
                            model[key] = value;
                        }
                    } else {
                        // drill down, creating missing structures on the way
                        if (typeof model[key] == 'string') delete model[key];
                        model = model[key] ? model[key] : (model[key] = array ? [] : {});
                        if (array) model = model[index] ? model[index] : (model[index] = {});
                    }
                }
            }
        }
    };
    sync.index = function(node) {
        var name = node.attr('name'),
            matches = name ? name.match(/:[0-9]+/g) : null;
        return matches ? matches[matches.length-1].substring(1) : node.index();
    };
    sync.match = function(name, names) {
        if (name in names) return names[name];// direct match
        if (name.indexOf(':') >= 0) {// check no indices match
            var noIndex = name.replace(/\:([0-9]+|i)/g, '');
            if (noIndex in names) return names[name] = names[noIndex];// cache result
        }
        for (var n in names) {//check for wildcard match
            if (n.indexOf('*') >= 0) {
                var re = new RegExp('^'+n.replace(/\./g, '\.').replace(/\*/g, '.*')+'$');
                if (re.test(noIndex||name)) return names[name] = names[n];// cache result
            }
        }
        return names[name] = null;// cache null result
    };

    // focusout events on editable elements trigger sync events on [name] parent/self element(s)
    sync.editable = 'input:not(:checkbox),select,textarea,[contenteditable="true"]';
    sync.onEvent = function(e) {
        $(e.target).parents().andSelf().filter('[sync],[sync=]').trigger('sync', [e]);
    };
    $(document)
    .delegate(sync.editable, 'focusout.sync', sync.onEvent)
    .delegate(':checkbox', 'click.sync', sync.onEvent)// checkbox focus is inconsistent across browsers
    .bind('sync.sync', function(e) {
        var node = $(e.target),
            attr = node.attr('sync'),
            name = attr ? attr : node.attr('name');
        if (!name) {
            node.trigger('focusout.sync');
        } else {
            var read = sync.match(name, sync.read) || sync.read;// watch for custom reader
            sync.set(name, read(node), e.target);
        }
    });

})(jQuery, document, undefined);
